import unittest
from unittest.mock import Mock, patch

import pyrefinebio
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

gorilla = {
    "name": "GORILLA",
    "taxonomy_id": 9592,
    "has_compendia": True,
    "has_quantfile_compendia": False,
}

mouse = {
    "name": "MUS",
    "taxonomy_id": 862507,
    "has_compendia": False,
    "has_quantfile_compendia": True,
}

search_1 = {"count": 2, "next": "search_2", "previous": None, "results": [gorilla]}

search_2 = {"count": 2, "next": None, "previous": "search_1", "results": [mouse]}


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/organisms/GORILLA/":
        return MockResponse(gorilla, url)

    if url == "https://api.refine.bio/v1/organisms/HUMAN/":
        return MockResponse(None, url, status=404)

    if url == "https://api.refine.bio/v1/organisms/500/":
        return MockResponse(None, url, status=500)

    if url == "https://api.refine.bio/v1/organisms/":
        return MockResponse(search_1, "search_2")

    if url == "search_2":
        return MockResponse(search_2, url)


class OrganismTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_organism_get(self, mock_request):
        result = pyrefinebio.Organism.get("GORILLA")
        self.assertObject(result, gorilla)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_organism_500(self, mock_request):
        with self.assertRaises(Exception):
            pyrefinebio.Organism.get(500)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_organism_get_404(self, mock_request):
        with self.assertRaises(Exception):
            pyrefinebio.Organism.get("HUMAN")

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_organism_search_no_filters(self, mock_request):
        results = pyrefinebio.Organism.search()

        self.assertObject(results[0], gorilla)
        self.assertObject(results[1], mouse)

    def test_organism_search_with_filters(self):
        filtered_results = pyrefinebio.Organism.search(has_quantfile_compendia=True)

        for result in filtered_results:
            self.assertTrue(result.has_quantfile_compendia)

    def test_organism_search_with_invalid_filters(self):
        with self.assertRaises(pyrefinebio.exceptions.InvalidFilters):
            pyrefinebio.OriginalFile.search(foo="bar")



from pyrefinebio import annotation as prb_annotation, sample as prb_sample
from pyrefinebio.api_interface import get_by_endpoint
from pyrefinebio.base import Base
from pyrefinebio.util import create_paginated_list, parse_date


class Experiment(Base):
    """Experiment

    Retrieve an Experiment based on id

        >>> import pyrefinebio
        >>> accession_code = 1
        >>> experiment = pyrefinebio.Experiment.get(accession_code)

    Search for Experiments based on filters

        >>> import pyrefinebio
        >>> experiments = pyrefinebio.Experiment.search(is_compendia=True, is_public=True)
    """

    def __init__(
        self,
        id=None,
        title=None,
        description=None,
        technology=None,
        annotations=None,
        samples=None,
        protocol_description=None,
        accession_code=None,
        alternate_accession_code=None,
        source_database=None,
        source_url=None,
        has_publication=None,
        publication_title=None,
        publication_doi=None,
        publication_authors=None,
        pubmed_id=None,
        source_first_published=None,
        source_last_modified=None,
        submitter_institution=None,
        platform_names=None,
        platform_accession_codes=None,
        last_modified=None,
        created_at=None,
        organism_names=None,
        downloadable_organism_names=None,
        sample_metadata_fields=None,
        sample_metadata=None,
        num_total_samples=None,
        num_processed_samples=None,
        num_downloadable_samples=None,
    ):
        super().__init__(identifier=accession_code)

        self.id = id
        self.title = title
        self.description = description
        self.technology = technology
        self.annotations = (
            [prb_annotation.Annotation(**a) for a in annotations] if annotations else []
        )
        self.samples = [prb_sample.Sample(**sample) for sample in samples] if samples else []
        self.protocol_description = protocol_description
        self.accession_code = accession_code
        self.alternate_accession_code = alternate_accession_code
        self.source_database = source_database
        self.source_url = source_url
        self.has_publication = has_publication
        self.publication_title = publication_title
        self.publication_doi = publication_doi
        self.publication_authors = publication_authors
        self.pubmed_id = pubmed_id
        self.source_first_published = parse_date(source_first_published)
        self.source_last_modified = parse_date(source_last_modified)
        self.submitter_institution = submitter_institution
        self.platform_names = platform_names
        self.platform_accession_codes = platform_accession_codes
        self.last_modified = parse_date(last_modified)
        self.created_at = parse_date(created_at)
        self.organism_names = organism_names
        self.downloadable_organism_names = downloadable_organism_names
        self.sample_metadata_fields = sample_metadata_fields
        self.sample_metadata = sample_metadata
        self.num_total_samples = num_total_samples
        self.num_processed_samples = num_processed_samples
        self.num_downloadable_samples = num_downloadable_samples

    @classmethod
    def get(cls, accession_code):
        """Retrieve an Experiment based on accession code

        Returns:
            Experiment

        Parameters:
            accession code (str): the accession code for the Experiment you want to get
        """
        response = get_by_endpoint("experiments/" + accession_code).json()
        return Experiment(**response)

    @classmethod
    def search(cls, **kwargs):
        """Search for Experiments based on various filters

        Returns:
            list of Experiment

        Keyword Arguments:
            id (int): filter based on the id of the Experiment

            technology (str): filter based on the technology used for the Experiment
                              (microarray, rna-seq, etc) this filter can have multiple values.

            has_publication (bool): filter based on if the Experiment has associated publications

            accession_code (str): filter based on the Experiment's accession code
                                  this filter can have multiple values

            alternate_accession_code (str): filter based on the Experiment's alternate
                                            accession codes

            platform (str): filter based on  platform, this filter can have multiple values

            organism (str): filter based on Organism, this filter can have multiple values

            downloadable_organism (str): filter based on Organisms that have downloadable files,
                                         this filter can have multiple values

            num_processed_samples (number): filter based on the Experiment's number of processed samples

            num_downloadable_samples (int): filter based on the Experiment's number of downloadable samples

            sample_keywords (str): filter based on keywords associated with the Experiment's Samples

            ordering (str): which field from to use when ordering the results

            search (str): specify a keyword which will be applied to the Experiment's title, publication_authors,
                          sample_keywords, publication_title, submitter_institution, description, accession_code,
                          alternate_accession_code, publication_doi, pubmed_id, sample_metadata_fields, and
                          platform_names

            limit (int): number of results to return per page

            offset (int): the initial index from which to return the results
        """
        response = get_by_endpoint("search", params=kwargs)
        return create_paginated_list(cls, response)



import unittest
from unittest.mock import Mock, patch

from click.testing import CliRunner
from pyrefinebio.script import cli


class ScriptTests(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()

    @patch("pyrefinebio.script.hlf.download_dataset")
    def test_download_dataset_dataset_dict(self, mock_download_dataset):
        self.runner.invoke(
            cli,
            [
                "download-dataset",
                "--email-address",
                "foo@bar.net",
                "--path",
                "./test.zip",
                "--dataset-dict",
                '{"foo": ["bar"]}',
            ],
        )

        mock_download_dataset.assert_called_with(
            "./test.zip",
            "foo@bar.net",
            {"foo": ["bar"]},
            None,
            "EXPERIMENT",
            "NONE",
            False,
            timeout=None,
            notify_me=False,
        )

    @patch("pyrefinebio.script.hlf.download_dataset")
    def test_download_dataset_notify_me(self, mock_download_dataset):
        self.runner.invoke(
            cli,
            [
                "download-dataset",
                "--email-address",
                "foo@bar.net",
                "--path",
                "./test.zip",
                "--dataset-dict",
                '{"foo": ["bar"]}',
                "--notify-me",
            ],
        )

        mock_download_dataset.assert_called_with(
            "./test.zip",
            "foo@bar.net",
            {"foo": ["bar"]},
            None,
            "EXPERIMENT",
            "NONE",
            False,
            timeout=None,
            notify_me=True,
        )

    @patch("pyrefinebio.script.hlf.download_dataset")
    def test_download_dataset_experiments(self, mock_download_dataset):
        self.runner.invoke(
            cli,
            [
                "download-dataset",
                "--email-address",
                "foo@bar.net",
                "--path",
                "./test.zip",
                "--experiments",
                "foo bar baz",
            ],
        )

        mock_download_dataset.assert_called_with(
            "./test.zip",
            "foo@bar.net",
            None,
            ["foo", "bar", "baz"],
            "EXPERIMENT",
            "NONE",
            False,
            timeout=None,
            notify_me=False,
        )

    @patch("pyrefinebio.script.hlf.download_dataset")
    def test_download_dataset_defualt_path(self, mock_download_dataset):
        self.runner.invoke(
            cli,
            ["download-dataset", "--email-address", "foo@bar.net", "--experiments", "foo bar baz"],
        )

        mock_download_dataset.assert_called_with(
            "./",
            "foo@bar.net",
            None,
            ["foo", "bar", "baz"],
            "EXPERIMENT",
            "NONE",
            False,
            timeout=None,
            notify_me=False,
        )

    def test_download_dataset_both(self):
        result = self.runner.invoke(
            cli,
            [
                "download-dataset",
                "--email-address",
                "foo@bar.net",
                "--path",
                "./test.zip",
                "--dataset-dict",
                '{"foo": ["bar"]}',
                "--experiments",
                "foo bar baz",
            ],
        )

        self.assertEqual(result.exit_code, 1)
        self.assertEqual(
            result.output,
            "Error: Unable to download Dataset\nYou should either provide dataset_dict or experiments but not both\n",
        )

    @patch("pyrefinebio.script.hlf.download_compendium")
    def test_download_compendium(self, mock_download_compendium):
        self.runner.invoke(
            cli,
            [
                "download-compendium",
                "--organism",
                "foo",
                "--path",
                "./test.zip",
                "--quant-sf-only",
                "true",
            ],
        )

        mock_download_compendium.assert_called_with("./test.zip", "foo", True)

    @patch("pyrefinebio.script.hlf.download_quantfile_compendium")
    def test_download_quantpendium(self, mock_download_q_compendium):
        self.runner.invoke(
            cli, ["download-quantfile-compendium", "--organism", "foo", "--path", "./test.zip"]
        )

        mock_download_q_compendium.assert_called_with("./test.zip", "foo")

    def test_download_compendium_bad_organism(self):
        result = self.runner.invoke(
            cli, ["download-compendium", "--organism", "foo", "--path", "./test.zip"]
        )

        self.assertEqual(result.exit_code, 1)
        self.assertEqual(
            result.output,
            "Searching for Compendium...\n"
            "Error: Unable to download Compendium\nCould not find any Compendium "
            + "with organism name, foo, version False, and quant_sf_only, False\n",
        )

    @patch("pyrefinebio.script.hlf.create_token")
    def test_create_token_prompt(self, mock_create_token):
        self.runner.invoke(cli, ["create-token"])

        mock_create_token.assert_called_with(None, None)

    @patch("pyrefinebio.script.hlf.create_token")
    def test_create_token_automatic(self, mock_create_token):
        self.runner.invoke(cli, ["create-token", "-s"])

        mock_create_token.assert_called_with(True, True)



from pyrefinebio import original_file as prb_original_file
from pyrefinebio.api_interface import get_by_endpoint
from pyrefinebio.base import Base
from pyrefinebio.util import create_paginated_list, parse_date


class DownloaderJob(Base):
    """DownloaderJob.

    Retrieve a DownloaderJob by id

        >>> import pyrefinebio
        >>> id = 1
        >>> job = pyrefinebio.DownloaderJob.get(id)

    Retrieve a list of DownloaderJobs based on filters

        >>> import pyrefinebio
        >>> jobs = pyrefinebio.DownloaderJob.search(batch_job_id=0, success=True)
    """

    def __init__(
        self,
        id=None,
        downloader_task=None,
        num_retries=None,
        retried=None,
        was_recreated=None,
        worker_id=None,
        ram_amount=None,
        worker_version=None,
        batch_job_id=None,
        batch_job_queue=None,
        failure_reason=None,
        success=None,
        original_files=[],
        start_time=None,
        end_time=None,
        created_at=None,
        last_modified=None,
        is_queued=None,
    ):
        super().__init__(identifier=id)

        self.id = id
        self.downloader_task = downloader_task
        self.num_retries = num_retries
        self.retried = retried
        self.was_recreated = was_recreated
        self.worker_id = worker_id
        self.ram_amount = ram_amount
        self.worker_version = worker_version
        self.batch_job_id = batch_job_id
        self.batch_job_queue = batch_job_queue
        self.failure_reason = failure_reason
        self.success = success
        self.start_time = parse_date(start_time)
        self.end_time = parse_date(end_time)
        self.created_at = parse_date(created_at)
        self.last_modified = parse_date(last_modified)
        self.is_queued = is_queued

        original_files_list = []
        if original_files:
            for original_file in original_files:
                if type(original_file) is dict:
                    original_files_list.append(prb_original_file.OriginalFile(**original_file))
                elif type(original_file) is int:
                    original_files_list.append(prb_original_file.OriginalFile(id=original_file))
                else:
                    original_files_list.append(original_file)

        self.original_files = original_files_list

    @classmethod
    def get(cls, id):
        """Retrieve a DownloaderJob based on id

        Returns:
            DownloaderJob

        Parameters:
            id (int): the id for the DownloaderJob you want to get
        """
        response = get_by_endpoint("jobs/downloader/" + str(id)).json()
        return DownloaderJob(**response)

    @classmethod
    def search(cls, **kwargs):
        """Retrieve a list of DownloaderJobs based on various filters

        Returns:
            list of DownloaderJob

        Keyword Arguments:

            id (int): filter based on the id of the DownloaderJob

            downloader_task (str): filter based on the job's task type

            num_retries (int): filter based on the number of times the job has retried

            retried (bool): filter based on if the job has retried

            was_recreated (bool): filter based on if the job was recreated

            worker_id (str): filter based on the job's worker id

            ram_amount (int): filter based on the amount of ram assigned to the job

            worker_version (str): filter based on the job's worker version

            batch_job_id (str): filter based on the job's batch id

            failure_reason (str): filter based on the reason why the job failed

            success (bool): filter based on if the job succeeded

            original_files (str): filter based on the ids of the OriginalFiles associated with the job

            start_time (str): filter based on the time when the job started

            end_time (str): filter based on the time when the job finished

            created_at (str): filter based on the time that the job was created

            last_modified (str): filter based on the time that the job was last modified

            ordering (str): which field to use when ordering the results.

            limit (int): number of results to return per page.

            offset (int): the initial index from which to return the results.

            sample_accession_code (str): filter based on the Samples associated with the job
        """
        response = get_by_endpoint("jobs/downloader", params=kwargs)
        return create_paginated_list(cls, response)


class ProcessorJob(Base):
    """Processor Job.

    Retrieve a ProcessorJob by id

        >>> import pyrefinebio
        >>> id = 1
        >>> job = pyrefinebio.ProcessorJob.get(id)

    Retrieve a list of ProcessorJobs based on filters

        >>> import pyrefinebio
        >>> jobs = pyrefinebio.ProcessorJob.search(num_retries=1)
    """

    def __init__(
        self,
        id=None,
        pipeline_applied=None,
        num_retries=None,
        retried=None,
        worker_id=None,
        ram_amount=None,
        volume_index=None,
        worker_version=None,
        failure_reason=None,
        batch_job_id=None,
        batch_job_queue=None,
        success=None,
        original_files=[],
        datasets=None,
        start_time=None,
        end_time=None,
        created_at=None,
        last_modified=None,
        is_queued=None,
    ):
        super().__init__(identifier=id)

        self.id = id
        self.pipeline_applied = pipeline_applied
        self.num_retries = num_retries
        self.retried = retried
        self.worker_id = worker_id
        self.ram_amount = ram_amount
        self.volume_index = volume_index
        self.worker_version = worker_version
        self.failure_reason = failure_reason
        self.batch_job_id = batch_job_id
        self.batch_job_queue = batch_job_queue
        self.success = success
        self.datasets = datasets
        self.start_time = parse_date(start_time)
        self.end_time = parse_date(end_time)
        self.created_at = parse_date(created_at)
        self.last_modified = parse_date(last_modified)
        self.is_queued = is_queued

        original_files_list = []
        if original_files:
            for original_file in original_files:
                if type(original_file) is dict:
                    original_files_list.append(prb_original_file.OriginalFile(**original_file))
                elif type(original_file) is int:
                    original_files_list.append(prb_original_file.OriginalFile(id=original_file))
                else:
                    original_files_list.append(original_file)

        self.original_files = original_files_list

    @classmethod
    def get(cls, id):
        """Retrieve a ProcessorJob based on id

        Returns:
            ProcessorJob

        Parameters:
            id (int): the id for the ProcessorJob you want to get
        """
        response = get_by_endpoint("jobs/processor/" + str(id)).json()
        return ProcessorJob(**response)

    @classmethod
    def search(cls, **kwargs):
        """Retrieve a list of ProcessorJobs based on various filters

        Returns:
            list of ProcessorJob

        Keyword Arguments:
            id (int): filter based on the id of the ProcessorJob

            pipeline_applied (str): filter based on the type of pipeline applied to the job

            num_retries (int): filter based on the number of times the job has retried

            retried (bool): filter based on if the job has retried

            worker_id (str): filter based on the id of the job's worker

            ram_amount (int): filter based on the amount of ram assigned to the job

            volume_index (str):

            worker_version (str): filter based on the job's worker version

            failure_reason (str): filter based on the reason why the job failed

            batch_job_id (str): filter based on the job's batch id

            success (bool): filter based on if the job has succeeded

            original_files (str): filter based on the ids of the OriginalFiles associated with the job

            datasets (str): filter based on the ids of the Datasets associated with the job

            start_time (str): filter based on the time when the job started

            end_time (str): filter based on the time when the job finished

            created_at (str): filter based on the time when the job was created

            last_modified (str): filter based on the time when the job was last modified

            ordering (str): which field to use when ordering the results.

            limit (int): number of results to return per page.

            offset (int): the initial index from which to return the results.

            sample_accession_code (str): filter based on the samples associated with the job
        """
        response = get_by_endpoint("jobs/processor", params=kwargs)
        return create_paginated_list(cls, response)


class SurveyJob(Base):
    """Survey Job.

    Retrieve a SurveyJob by id

        >>> import pyrefinebio
        >>> id = 1
        >>> job = pyrefinebio.SurveyJob.get(id)

    Retrieve a list of SurveyJobs based on filters

        >>> import pyrefinebio
        >>> jobs = pyrefinebio.SurveyJob.search(num_retries=1)
    """

    def __init__(
        self,
        id=None,
        source_type=None,
        success=None,
        ram_amount=None,
        start_time=None,
        end_time=None,
        batch_job_id=None,
        batch_job_queue=None,
        created_at=None,
        last_modified=None,
        is_queued=None,
    ):
        super().__init__(identifier=id)

        self.id = id
        self.source_type = source_type
        self.success = success
        self.ram_amount = ram_amount
        self.start_time = parse_date(start_time)
        self.end_time = parse_date(end_time)
        self.batch_job_id = batch_job_id
        self.batch_job_queue = batch_job_queue
        self.created_at = parse_date(created_at)
        self.last_modified = parse_date(last_modified)
        self.is_queued = is_queued

    @classmethod
    def get(cls, id):
        """Retrieve a SurveyJob based on id

        Returns:
            SurveyJob

        Parameters:
            id (int): the id for the SurveyJob you want to get
        """
        response = get_by_endpoint("jobs/survey/" + str(id)).json()
        return SurveyJob(**response)

    @classmethod
    def search(cls, **kwargs):
        """Retrieve a list of SurveyJobs based on various filters

        Returns:
            list of SurveyJob

        Keyword Arguments:
            id (int): filter based on the id of the SurveyJob

            source_type (str): filter based on the name of the source database

            success (bool): filter based on if the job has succeeded

            ram_amount (int): filter based on the amount of ram assigned to the job

            start_time (str): filter based on the time when the job started

            end_time (str): filter based on the time when the job finished

            created_at (str): filter based on the time when the job was created

            last_modified (str): filter based on the time when the job was last modified

            ordering (str): which field to use when ordering the results.

            limit (int): number of results to return per page.

            offset (int): the initial index from which to return the results.
        """
        response = get_by_endpoint("jobs/survey", params=kwargs)
        return create_paginated_list(cls, response)



import os
import unittest
from unittest.mock import Mock, patch

import pyrefinebio
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

compendium_object_1 = {
    "id": 1,
    "primary_organism_name": "HUMAN",
    "organism_names": ["HUMAN", "GORILLA"],
    "svd_algorithm": "NONE",
    "quant_sf_only": True,
    "compendium_version": 1,
    "computed_file": {
        "id": 420,
        "filename": "test-compendium.zip",
        "size_in_bytes": 5000,
        "is_smashable": False,
        "is_qc": False,
        "sha1": "3c5de2dba42c2768e7542a0262d2b9a1e9fc9f45",
        "s3_bucket": "test-bucket",
        "s3_key": "test-compendium.zip",
        "download_url": "test_download_url",
        "created_at": "2020-10-02T00:00:00Z",
        "last_modified": "2020-10-02T00:00:00Z",
    },
}

compendium_object_2 = {
    "id": 42,
    "primary_organism_name": "GORILLA",
    "organism_names": ["HUMAN", "GORILLA"],
    "svd_algorithm": "NONE",
    "quant_sf_only": True,
    "compendium_version": 1,
    "computed_file": {
        "id": 4,
        "filename": "test-compendium-2.zip",
        "size_in_bytes": 7500,
        "is_smashable": True,
        "is_qc": False,
        "sha1": "3c5de2dba42c2768e7542a0262d2b9a1e9fc9f45",
        "s3_bucket": "test-bucket",
        "s3_key": "test-compendium-2.zip",
        "created_at": "2020-10-05T00:00:00Z",
        "last_modified": "2020-10-05T00:00:00Z",
    },
}

search_1 = {"count": 2, "next": "search_2", "previous": None, "results": [compendium_object_1]}

search_2 = {"count": 2, "next": None, "previous": "search_1", "results": [compendium_object_2]}


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/compendia/1/":
        return MockResponse(compendium_object_1, url)

    if url == "https://api.refine.bio/v1/compendia/42/":
        return MockResponse(compendium_object_2, url)

    if url == "https://api.refine.bio/v1/compendia/0/":
        return MockResponse(None, url, status=404)

    if url == "https://api.refine.bio/v1/compendia/500/":
        return MockResponse(None, url, status=500)

    if url == "https://api.refine.bio/v1/compendia/":
        return MockResponse(search_1, "search_2")

    if url == "search_2":
        return MockResponse(search_2, url)


class CompendiumTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_compendium_get(self, mock_request):
        result = pyrefinebio.Compendium.get(1)
        self.assertObject(result, compendium_object_1)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_compendium_500(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.ServerError):
            pyrefinebio.Compendium.get(500)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_compendium_get_404(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.NotFound):
            pyrefinebio.Compendium.get(0)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_compendium_search_no_filters(self, mock_request):
        results = pyrefinebio.Compendium.search()

        self.assertObject(results[0], compendium_object_1)
        self.assertObject(results[1], compendium_object_2)

    def test_compendium_search_with_filters(self):
        filtered_results = pyrefinebio.Compendium.search(
            primary_organism__name="ACTINIDIA_CHINENSIS"
        )

        for result in filtered_results:
            self.assertEqual(result.primary_organism_name, "ACTINIDIA_CHINENSIS")

    def test_compendium_search_with_invalid_filters(self):
        with self.assertRaises(pyrefinebio.exceptions.InvalidFilters):
            pyrefinebio.Compendium.search(foo="bar")

    # just mock download - it's already tested in depth in test_dataset
    @patch("pyrefinebio.compendia.download_file")
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_compendium_download(self, mock_request, mock_download):
        result = pyrefinebio.Compendium.get(1)

        result.download("test-path")

        mock_download.assert_called_with("test_download_url", os.path.abspath("test-path"), True)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_compendium_download_no_url(self, mock_request):
        result = pyrefinebio.Compendium.get(42)  # 42 has no download_url

        result.computed_file._fetched = (
            True  # set fetched to true so that a request isn't made to /computed_files
        )

        with self.assertRaises(pyrefinebio.exceptions.DownloadError) as de:
            result.download("test-path")

    @patch("pyrefinebio.compendia.shutil.unpack_archive")
    def test_compendium_extract(self, mock_unpack):
        c = pyrefinebio.Compendium()

        c._downloaded_path = "foo"

        c.extract()

        mock_unpack.assert_called_with("foo")



from pyrefinebio import job as prb_job, sample as prb_sample
from pyrefinebio.api_interface import get_by_endpoint
from pyrefinebio.base import Base
from pyrefinebio.util import create_paginated_list, parse_date


class OriginalFile(Base):
    """Original File.

    Retrieve an OriginalFile by id

        >>> import pyrefinebio
        >>> id = 1
        >>> og_file = pyrefinebio.OriginalFile.get(id)

    Retrieve a list of OriginalFiles based on filters

        >>> import pyrefinebio
        >>> og_files = pyrefinebio.OriginalFile.search()
    """

    def __init__(
        self,
        id=None,
        filename=None,
        size_in_bytes=None,
        sha1=None,
        samples=None,
        processor_jobs=None,
        downloader_jobs=None,
        source_url=None,
        source_filename=None,
        is_downloaded=None,
        is_archive=None,
        has_raw=None,
        created_at=None,
        last_modified=None,
    ):
        super().__init__(identifier=id)

        self.id = id
        self.filename = filename
        self.size_in_bytes = size_in_bytes
        self.sha1 = sha1
        samples_list = []
        if samples:
            for sample in samples:
                if type(sample) is dict:
                    samples_list.append(prb_sample.Sample(**sample))
                else:
                    samples_list.append(sample)

        self.samples = samples_list

        processor_jobs_list = []
        if processor_jobs:
            for processor_job in processor_jobs:
                if type(processor_job) is dict:
                    processor_jobs_list.append(prb_job.ProcessorJob(**processor_job))
                else:
                    processor_jobs_list.append(processor_job)

        self.processor_jobs = processor_jobs_list

        downloader_jobs_list = []
        if downloader_jobs:
            for downloader_job in downloader_jobs:
                if type(downloader_job) is dict:
                    downloader_jobs_list.append(prb_job.DownloaderJob(**downloader_job))
                else:
                    downloader_jobs_list.append(downloader_job)

        self.downloader_jobs = downloader_jobs_list

        self.source_url = source_url
        self.source_filename = source_filename
        self.is_downloaded = is_downloaded
        self.is_archive = is_archive
        self.has_raw = has_raw
        self.created_at = parse_date(created_at)
        self.last_modified = parse_date(last_modified)

    @classmethod
    def get(cls, id):
        """Retrieve an OriginalFile based on id

        Returns:
            OriginalFile

        Parameters:
            id (int): the id for the OriginalFile you want to get
        """

        response = get_by_endpoint("original_files/" + str(id)).json()
        return OriginalFile(**response)

    @classmethod
    def search(cls, **kwargs):
        """Retrieve a list of OriginalFiles based on various filters

        Returns:
            list of OriginalFile

        Keyword Arguments:

            id (int): filter based on the id of the OriginalFile

            filename (str): filter based on the name of the OriginalFile

            samples (str): filter based on the Samples associated with the OriginalFile

            size_in_bytes (int): filter based on the OriginalFile's size

            sha1 (str): filter based on the OriginalFiles sha1 hash

            processor_jobs (str): filter based on the ProcessorJobs associated with the OriginalFile

            downloader_jobs (str): filter based on the DownloaderJobs associated with the OriginalFile

            source_url (str): filter based on the OriginalFile's source url

            is_archive (bool): filter based on if the OriginalFile is archived

            source_filename (str): filter based on the OriginalFile's source's filename

            has_raw (bool): filter based on if the OriginalFile had raw data available in the source
                            database

            created_at (str): filter based on the time when the OriginalFile was created

            last_modified (str): filter based on the time when the OriginalFile was last modified

            ordering (str): which field to use when ordering the results.

            limit (int): number of results to return per page.

            offset (int): the initial index from which to return the results.
        """

        response = get_by_endpoint("original_files", params=kwargs)
        return create_paginated_list(cls, response)



import json
import os
import unittest
from unittest.mock import Mock, patch

import pyrefinebio
from pyrefinebio.config import Config
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

dataset = {
    "id": "test-dataset",
    "data": {"test-experiment": ["sample-1", "sample-2"]},
    "aggregate_by": "EXPERIMENT",
    "scale_by": "NONE",
    "is_processing": True,
    "is_processed": True,
    "is_available": True,
    "has_email": True,
    "expires_on": "2020-10-29T16:12:56.449529Z",
    "s3_bucket": "test-bucket",
    "s3_key": "test-key",
    "success": True,
    "failure_reason": "",
    "created_at": "2020-10-19T19:30:36.157340Z",
    "last_modified": "2020-10-22T16:12:56.449544Z",
    "start": None,
    "size_in_bytes": 173109,
    "sha1": "test-sha1",
    "download_url": "test-dl-url",
    "quantile_normalize": True,
    "quant_sf_only": False,
    "svd_algorithm": "NONE",
}

non_processed_dataset = {
    "id": "non-processed-dataset",
    "is_processing": False,
    "is_processed": False,
    "download_url": None,
}


def mock_request(method, url, data, **kwargs):

    if method == "POST":
        data = json.loads(data)
        ds = {}
        for key, value in dataset.items():
            if data.get(key):
                ds[key] = data.get(key)
            else:
                ds[key] = value
        return MockResponse(ds, url)

    if url == "https://api.refine.bio/v1/dataset/test-dataset/":
        return MockResponse(dataset, url)

    if url == "https://api.refine.bio/v1/dataset/non-processed-dataset/":
        return MockResponse(non_processed_dataset, url)

    if url == "https://api.refine.bio/v1/dataset/":
        return MockResponse(dataset, url)

    if url == "https://api.refine.bio/v1/dataset/500/":
        return MockResponse(dataset, url, status=500)


class DatasetTests(unittest.TestCase, CustomAssertions):
    @classmethod
    def tearDownClass(cls):
        os.environ.pop("REFINEBIO_CONFIG_FILE", None)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_dataset_get(self, mock_request):
        result = pyrefinebio.Dataset.get("test-dataset")
        self.assertObject(result, dataset)

    def test_dataset_get_404(self):
        with self.assertRaises(pyrefinebio.exceptions.NotFound):
            pyrefinebio.Dataset.get("this-does-not-exist")

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_dataset_500(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.ServerError):
            pyrefinebio.Dataset.get("500")

    @patch("pyrefinebio.dataset.post_by_endpoint")
    @patch("pyrefinebio.dataset.put_by_endpoint")
    def test_dataset_save(self, mock_put, mock_post):
        mock_post.return_value = MockResponse(dataset, "")
        mock_put.return_value = MockResponse(dataset, "")

        ds = pyrefinebio.Dataset(
            data={"test-experiment": ["sample-1", "sample-2"]}, email_address="test-email"
        )
        ds = ds.save()

        mock_post.assert_called_with(
            "dataset",
            payload={
                "data": {"test-experiment": ["sample-1", "sample-2"]},
                "email_address": "test-email",
                "notify_me": False,
            },
        )

        ds.email_address = "changed-email"
        ds.save()

        mock_put.assert_called_with(
            "dataset/test-dataset",
            payload={
                "data": {"test-experiment": ["sample-1", "sample-2"]},
                "aggregate_by": "EXPERIMENT",
                "scale_by": "NONE",
                "email_address": "changed-email",
                "quantile_normalize": True,
                "quant_sf_only": False,
                "svd_algorithm": "NONE",
                # We want email notifications to be opt-out
                "notify_me": False,
            },
        )

    def test_dataset_save_bad_data(self):
        with self.assertRaises(pyrefinebio.exceptions.InvalidData):
            pyrefinebio.Dataset(data={"lol": ["not-good"]}).save()

    @patch("pyrefinebio.dataset.post_by_endpoint")
    def test_dataset_skip_quantile_normalization(self, mock_post_by_endpoint):
        """
        Test skipping quantile normalization

        Test the corner case brought up by Steven where we set
        quantile_normalize False. We used to check if quantile_normalize was
        truthy before adding it to the request, which caused it to always be
        true.
        """

        pyrefinebio.Dataset(
            data={"GSE68833": ["ALL"]},
            aggregate_by="EXPERIMENT",
            email_address="foo@bar.com",
            quantile_normalize=False,
        ).save()

        self.assertEqual(mock_post_by_endpoint.call_count, 1)

        # We need to use assertEqual here because None is False-y
        args, kwargs = mock_post_by_endpoint.call_args
        self.assertEqual(kwargs["payload"].get("quantile_normalize", None), False)

    @patch("pyrefinebio.dataset.post_by_endpoint")
    @patch("pyrefinebio.dataset.put_by_endpoint")
    def test_dataset_notify_me(self, mock_put_by_endpoint, mock_post_by_endpoint):
        """
        Test requesting to be notified
        """
        mock_post_by_endpoint.return_value = MockResponse(dataset, "")

        # First test that notify_me is false by default
        ds = pyrefinebio.Dataset(
            data={"GSE68833": ["ALL"]},
            aggregate_by="EXPERIMENT",
            email_address="foo@bar.com",
        )
        ds.save()

        self.assertEqual(mock_post_by_endpoint.call_count, 1)

        # We need to use assertEqual here because None is False-y
        args, kwargs = mock_post_by_endpoint.call_args
        self.assertEqual(kwargs["payload"].get("notify_me", None), False)

        self.assertIsNotNone(ds.id)

        # Now set notify_me true
        ds.notify_me = True
        ds.save()

        self.assertEqual(mock_put_by_endpoint.call_count, 1)

        args, kwargs = mock_put_by_endpoint.call_args
        self.assertTrue(kwargs["payload"].get("notify_me", None))

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_dataset_process(self, mock_request):
        ds = pyrefinebio.Dataset(
            data={"test-experiment": ["sample-1", "sample-2"]}, email_address="test-email"
        )
        ds.process()

        self.assertTrue(ds.start)
        self.assertTrue(ds.is_processing)

    def test_dataset_process_bad_token(self):
        pyrefinebio.config.Config().token = "this-is-a-bad-token"

        data = {}
        exs = pyrefinebio.Experiment.search(num_downloadable_samples=1)
        ex = next(exs)
        data[ex.accession_code] = ["ALL"]

        ds = pyrefinebio.Dataset(data=data)

        with self.assertRaises(pyrefinebio.exceptions.BadRequest) as br:
            ds.process()

        self.assertEqual(
            br.exception.base_message, "Bad Request: You must provide an active API token ID"
        )

    def test_dataset_process_no_email(self):
        try:
            token = pyrefinebio.Token(id="42240c84-b3d9-4f41-8001-6f40abce9d7d")
            token.agree_to_terms_and_conditions()
        except pyrefinebio.exceptions.NotFound:
            token = pyrefinebio.Token()
            token.agree_to_terms_and_conditions()

        data = {}
        exs = pyrefinebio.Experiment.search(num_downloadable_samples=1)
        ex = next(exs)
        data[ex.accession_code] = ["ALL"]

        ds = pyrefinebio.Dataset(data=data)

        with self.assertRaises(pyrefinebio.exceptions.BadRequest) as br:
            ds.process()

        self.assertEqual(
            br.exception.base_message, "Bad Request: You must provide an email address."
        )

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_dataset_check(self, mock_request):
        ds = pyrefinebio.Dataset(id="test-dataset")
        is_done = ds.check()

        self.assertTrue(is_done)
        self.assertTrue(ds.is_processing)
        self.assertTrue(ds.is_processed)

    @patch("pyrefinebio.api_interface.shutil.copyfileobj")
    @patch("pyrefinebio.api_interface.open")
    @patch("pyrefinebio.api_interface.requests.get")
    def test_dataset_download(self, mock_get, mock_open, mock_copy):
        ds = pyrefinebio.Dataset(download_url="test_download_url")

        mr = MockResponse(None, "test_download_url", headers={"content-length": 3})

        setattr(mr, "raw", "raw")

        mock_get.return_value.__enter__.return_value = mr
        mock_open.return_value.__enter__.return_value = "file"

        ds.download("test_path")

        mock_get.assert_called_with("test_download_url", stream=True)
        mock_open.assert_called_with(os.path.abspath("test_path"), "wb")
        mock_copy.assert_called_with("raw", "file")

    @patch("pyrefinebio.api_interface.shutil.copyfileobj")
    @patch("pyrefinebio.api_interface.input")
    @patch("pyrefinebio.api_interface.requests.get")
    def test_dataset_download_big_file(self, mock_get, mock_input, mock_copy):
        ds = pyrefinebio.Dataset(download_url="test_download_url")

        mr = MockResponse(None, "test_download_url", headers={"content-length": 1000000000000})

        setattr(mr, "raw", "raw")

        mock_get.return_value.__enter__.return_value = mr
        mock_input.return_value = "n"

        ds.download("test_path")

        mock_get.assert_called_with("test_download_url", stream=True)
        mock_copy.assert_not_called()

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_dataset_download_no_url(self, mock_request):
        ds = pyrefinebio.Dataset(id="non-processed-dataset", download_url=None)
        with self.assertRaises(pyrefinebio.exceptions.DownloadError) as de:
            ds.download("test")

        self.assertEqual(
            str(de.exception),
            "Unable to download Dataset\nDownload url not found - you must process the Dataset before downloading.",
        )

    @patch("pyrefinebio.api_interface.shutil.copyfileobj")
    @patch("pyrefinebio.api_interface.input")
    @patch("pyrefinebio.api_interface.requests.get")
    def test_dataset_download_no_size_header(self, mock_get, mock_input, mock_copy):
        ds = pyrefinebio.Dataset(download_url="test_download_url")

        mr = MockResponse(None, "test_download_url", headers={})

        setattr(mr, "raw", "raw")

        mock_get.return_value.__enter__.return_value = mr
        mock_input.return_value = "n"

        ds.download("test_path")

        mock_get.assert_called_with("test_download_url", stream=True)
        mock_copy.assert_not_called()

    def test_add_samples(self):
        dataset = pyrefinebio.Dataset()

        dataset.add_samples("test-experiment-1")

        experiment2 = pyrefinebio.Experiment(accession_code="test-experiment-2")

        dataset.add_samples(experiment2)

        sample1 = pyrefinebio.Sample(accession_code="sample1")
        sample2 = pyrefinebio.Sample(accession_code="sample2")

        dataset.add_samples("test-experiment-3", samples=[sample1, sample2, "sample3"])

        self.assertDictEqual(
            dataset.data,
            {
                "test-experiment-1": ["ALL"],
                "test-experiment-2": ["ALL"],
                "test-experiment-3": ["sample1", "sample2", "sample3"],
            },
        )

    @patch("pyrefinebio.dataset.shutil.unpack_archive")
    def test_dataset_extract(self, mock_unpack):
        ds = pyrefinebio.Dataset()

        ds._downloaded_path = "foo"

        ds.extract()

        mock_unpack.assert_called_with("foo")



import os

from pyrefinebio.api_interface import get_by_endpoint, post_by_endpoint, put_by_endpoint
from pyrefinebio.base import Base
from pyrefinebio.config import Config
from pyrefinebio.exceptions import BadRequest, NotFound

config = Config()


class Token(Base):
    """Handles the creation, activation, saving, and loading of refine.bio's api tokens.

    These tokens can be used in requests that provide urls to download computed files.

    Please review refine.bio's [Terms of Use](https://www.refine.bio/terms) and
    [Privacy Policy](https://www.refine.bio/privacy) before use of these tokens.

    Create a new Token

        >>> import pyrefinebio
        >>> token = pyrefinebio.Token(email_address="foo@bar.com")

    Create a Token and activate it

        >>> import pyrefinebio
        >>> token = pyrefinebio.Token(email_address="foo@bar.com")
        >>> token.agree_to_terms_and_conditions()

    Save a Token to the config file

        >>> import pyrefinebio
        >>> token = pyrefinebio.Token(email_address="foo@bar.com")
        >>> token.save_token()

    Load the Token that is currently set to Config

        >>> import pyrefinebio
        >>> token = pyrefinebio.Token.load_token()
    """

    def __init__(
        self, email_address=None, id=None,
    ):
        if id is None:
            response = post_by_endpoint("token", payload={"email_address": email_address}).json()
            self.id = response["id"]
        else:
            self.id = id

        self.email_address = email_address

    def agree_to_terms_and_conditions(self):
        """Activates a token.

        Tokens that are activated will be added to Config and can be used to make requests
        until pyrefinebio is re-imported.

        If you would like to continue using the same Token for other scripts that use pyrefinebio
        save the Token to the Config file using save_token().

        Activating a token indicates agreement with refine.bio's
        [Terms of Use](https://www.refine.bio/terms) and
        [Privacy Policy](https://www.refine.bio/privacy).
        """
        try:
            put_by_endpoint("token/" + str(self.id), payload={"is_activated": True})
        except NotFound:
            raise BadRequest(
                "Token with id '" + str(self.id) + "' does not exist in refine.bio. "
                "Please create a new token."
            )

        config.token = self.id

    def save_token(self):
        """Saves a token to the config file.

        The default config file is ~/.refinebio.yaml, but
        you can use the environment variable `CONFIG_FILE` to change this path
        """
        try:
            response = get_by_endpoint("token/" + str(self.id)).json()
            if not response["is_activated"]:
                raise BadRequest(
                    "Token with id '" + str(self.id) + "' is not activated. "
                    "Please activate your token with `agree_to_terms_and_conditions()` before saving it."
                )
        except NotFound:
            raise BadRequest(
                "Token with id '" + str(self.id) + "' does not exist in refine.bio. "
                "Please create a new token using pyrefinebio.Token()."
            )

        config.token = self.id
        config.save()

    @classmethod
    def load_token(cls):
        """Loads the token that's currently set to Config."""
        return Token(id=config.token)



import json
import shutil

import requests
from pyrate_limiter import Duration, Limiter, RequestRate
from requests.exceptions import ConnectionError

from pyrefinebio.config import Config
from pyrefinebio.exceptions import (
    BadRequest,
    InvalidData,
    InvalidFilters,
    InvalidFilterType,
    MultipleErrors,
    NotFound,
    ServerError,
)

CONFIG = Config()

# Rate limit the API requests per second.
limiter = Limiter(RequestRate(CONFIG.api_max_calls_per_second, Duration.SECOND))


@limiter.ratelimit("refinebio", delay=True)
def request(method, url, params=None, payload=None):
    try:
        headers = {"Content-Type": "application/json", "API-KEY": CONFIG.token}

        if payload:
            payload = json.dumps(payload)

        response = requests.request(method, url, params=params, data=payload, headers=headers)
        response.raise_for_status()

        return response

    except requests.exceptions.HTTPError as e:
        code = response.status_code
        try:
            response_body = response.json()
        except json.decoder.JSONDecodeError:
            response_body = response.text

        if code == 400:
            error = _handle_error(response_body)
            raise error

        elif code == 404:
            raise NotFound(response.url)

        elif code >= 500:
            raise ServerError()

        else:
            print(response_body)
            raise e
    except ConnectionError:
        raise ServerError()


def get(url, params=None):
    return request("GET", url, params=params)


def post(url, payload=None):
    return request("POST", url, payload=payload)


def put(url, payload=None):
    return request("PUT", url, payload=payload)


def get_by_endpoint(endpoint, params=None):
    config = Config()
    url = config.base_url + endpoint + "/"
    return get(url, params=params)


def post_by_endpoint(endpoint, payload=None):
    config = Config()
    url = config.base_url + endpoint + "/"
    return post(url, payload=payload)


def put_by_endpoint(endpoint, payload=None):
    config = Config()
    url = config.base_url + endpoint + "/"
    return put(url, payload=payload)


def download_file(url, path, prompt):
    with requests.get(url, stream=True) as res:

        if prompt:
            total_size_in_bytes = int(res.headers.get("content-length", -1))
            message = None

            if total_size_in_bytes == -1:
                message = (
                    "Could not get the size for the file you are tying to download. "
                    "Would you still like to download it? (y/N)"
                )
            if total_size_in_bytes > 1024 * 1024 * 1024:
                message = (
                    "The file you are trying to download is bigger than 1GB ({:.2f}GB). "
                    "Would you still like to download it? (y/N)"
                ).format(total_size_in_bytes / (1024 * 1024 * 1024))

            if message:
                yn = input(message)

                if yn.lower() not in ("y", "yes"):
                    return

        with open(path, "wb") as f:
            shutil.copyfileobj(res.raw, f)


def _handle_error(response_body):
    try:
        error_type = response_body["error_type"]
        message = response_body["message"]
        details = response_body["details"]

        if error_type == "multiple_errors":
            return _hanlde_multiple_errors(details)

        if error_type == "invalid_filters":
            return InvalidFilters(details)

        if error_type == "invalid_data":
            return InvalidData(message, details)

        if error_type == "invalid" or error_type == "invalid_choice":
            return InvalidFilterType(details, message)

        return BadRequest(message)

    except (ValueError, KeyError, TypeError):
        return BadRequest(response_body)


def _hanlde_multiple_errors(errors):
    exceptions = []

    for error in errors:
        exceptions.append(_handle_error(error))

    return MultipleErrors(exceptions)



import os

import setuptools

with open("README.md", "r") as readme:
    long_description = readme.read()

setuptools.setup(
    name="pyrefinebio",
    version=os.getenv("VERSION"),
    author="Childhood Cancer Data Lab",
    author_email="ccdl@alexslemonade.org",
    description="A python client for the refine.bio API.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AlexsLemonade/refinebio-py",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",  # not sure if this is actually true - we should probably test on windows!
    ],
    python_requires=">=3.6",
    install_requires=["iso8601", "PyYAML", "requests", "Click", "pytimeparse"],
    entry_points="""
        [console_scripts]
        refinebio=pyrefinebio.script:cli
    """,
)



import unittest
from copy import deepcopy
from unittest.mock import patch

import pyrefinebio
from pyrefinebio import original_file as prb_original_file
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

job_1 = {
    "id": 1,
    "downloader_task": "SRA",
    "num_retries": 0,
    "retried": False,
    "was_recreated": True,
    "worker_id": "i-013f4473d78ce5cef",
    "worker_version": "v1.39.3-dev-hotfix",
    "batch_job_id": "faf15c0e-3164-4bae-98f6-c6668f9ff4a5",
    "batch_job_queue": "data-refinery-batch-workers-queue-circleci-prod-0",
    "failure_reason": None,
    "success": True,
    "original_files": [412392, 412403],
    "start_time": "2020-09-29T21:34:43.735081Z",
    "end_time": "2020-09-29T21:34:57.725233Z",
    "created_at": "2020-07-01T16:47:18.384967Z",
    "last_modified": "2020-09-29T21:34:57.725247Z",
}

job_2 = {
    "id": 2,
    "downloader_task": "SRA",
    "num_retries": 2,
    "retried": True,
    "was_recreated": False,
    "worker_id": "i-097213f99819c2edc",
    "worker_version": "v1.36.9-dev-hotfix",
    "batch_job_id": "2d4174c9-c449-49c5-b9c2-f270ec4597f0",
    "batch_job_queue": "data-refinery-batch-workers-queue-circleci-prod-0",
    "failure_reason": "Exception caught while downloading file\\n [Errno 28] No space left on device",
    "success": False,
    "original_files": [396995],
    "start_time": "2020-06-25T09:56:06.692840Z",
    "end_time": "2020-06-25T10:07:55.920347Z",
    "created_at": "2020-06-25T09:56:04.731805Z",
    "last_modified": "2020-06-25T10:07:56.601791Z",
}

job_1_object_dict = deepcopy(job_1)

job_1_object_dict["original_files"] = [
    prb_original_file.OriginalFile(file_id) for file_id in job_1["original_files"]
]

job_2_object_dict = deepcopy(job_2)

job_2_object_dict["original_files"] = [
    prb_original_file.OriginalFile(file_id) for file_id in job_2["original_files"]
]

search_1 = {"count": 2, "next": "search_2", "previous": None, "results": [job_1]}

search_2 = {"count": 2, "next": None, "previous": "search_1", "results": [job_2]}


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/jobs/downloader/1/":
        return MockResponse(job_1, url)

    if url == "https://api.refine.bio/v1/jobs/downloader/2/":
        return MockResponse(job_2, url)

    if url == "https://api.refine.bio/v1/jobs/downloader/0/":
        return MockResponse(None, url, status=404)

    if url == "https://api.refine.bio/v1/jobs/downloader/500/":
        return MockResponse(None, url, status=500)

    if url == "https://api.refine.bio/v1/jobs/downloader/":
        return MockResponse(search_1, "search_2")

    if url == "search_2":
        return MockResponse(search_2, url)


class DownloaderJobTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_downloader_job_get(self, mock_request):
        result = pyrefinebio.DownloaderJob.get(1)
        self.assertObject(result, job_1_object_dict)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_downloader_job_500(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.ServerError):
            pyrefinebio.DownloaderJob.get(500)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_downloader_job_get_404(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.NotFound):
            pyrefinebio.DownloaderJob.get(0)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_downloader_job_search_no_filters(self, mock_request):
        results = pyrefinebio.DownloaderJob.search()

        self.assertObject(results[0], job_1_object_dict)
        self.assertObject(results[1], job_2_object_dict)

    def test_downloader_job_search_with_filters(self):
        filtered_results = pyrefinebio.DownloaderJob.search(
            success=True, retried=True, num_retries=1, downloader_task="GEO"
        )

        for result in filtered_results:
            self.assertTrue(result.success)
            self.assertTrue(result.retried)
            self.assertEqual(result.num_retries, 1)
            self.assertEqual(result.downloader_task, "GEO")

    def test_downloader_job_search_with_invalid_filters(self):
        with self.assertRaises(pyrefinebio.exceptions.InvalidFilters):
            pyrefinebio.DownloaderJob.search(foo="bar")



from pyrefinebio.compendia import Compendium
from pyrefinebio.computational_result import ComputationalResult
from pyrefinebio.computed_file import ComputedFile
from pyrefinebio.dataset import Dataset
from pyrefinebio.experiment import Experiment
from pyrefinebio.institution import Institution
from pyrefinebio.job import DownloaderJob, ProcessorJob, SurveyJob
from pyrefinebio.organism import Organism
from pyrefinebio.original_file import OriginalFile
from pyrefinebio.platform import Platform
from pyrefinebio.processor import Processor
from pyrefinebio.qn_target import QNTarget
from pyrefinebio.sample import Sample
from pyrefinebio.token import Token
from pyrefinebio.transcriptome_index import TranscriptomeIndex

from pyrefinebio.exceptions import (
    ServerError,
    BadRequest,
    InvalidFilters,
    InvalidFilterType,
    InvalidData,
    NotFound,
    DownloadError,
    MultipleErrors
)

from pyrefinebio.high_level_functions import (
    help,
    download_dataset,
    download_compendium,
    download_quantfile_compendium,
    create_token
)



import threading
import time
import unittest
from unittest.mock import patch

import pyrefinebio
from pyrefinebio.exceptions import BadRequest, NotFound
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse


def mock_200_request(method, url, **kwargs):
    experiment = {
        "title": "TEST_EXPERIMENT",
    }
    sample = {
        "accession_code": "TEST_SAMPLE",
    }

    if url == "https://api.refine.bio/v1/experiments/UNDER/":
        experiment["samples"] = [{"accession_code": str(i)} for i in range(9)]
        return MockResponse(experiment, url)

    if url == "https://api.refine.bio/v1/experiments/OVER/":
        experiment["samples"] = [{"accession_code": str(i)} for i in range(10)]
        return MockResponse(experiment, url)

    if url.startswith("https://api.refine.bio/v1/samples/"):
        return MockResponse(sample, url, 200)


def mock_400_request(method, url, **kwargs):
    return MockResponse(
        "we're not home right now", "https://api.refine.bio/v1/organisms/GORILLA", 400
    )


def mock_404_request(method, url, **kwargs):
    return MockResponse(
        "we're not home right now", "https://api.refine.bio/v1/organisms/GORILLA", 404
    )


class ApiInterfaceTests(unittest.TestCase, CustomAssertions):
    PERIOD_SECONDS = 1

    class Timer:
        def __enter__(self):
            self.start = time.time()
            return self

        def __exit__(self, *args, **kwargs):
            self.duration = time.time() - self.start

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_400_request)
    def test_400(self, mock_request):
        with self.assertRaises(BadRequest):
            pyrefinebio.Organism.get("GORILLA")

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_404_request)
    def test_404(self, mock_request):
        with self.assertRaises(NotFound):
            pyrefinebio.Organism.get("GORILLA")

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_200_request)
    def test_rate_limit(self, mock_request):
        time.sleep(self.PERIOD_SECONDS)
        threads = []
        with self.Timer() as timer:
            for sample in pyrefinebio.Experiment.get("UNDER").samples:
                t = threading.Thread(target=lambda: sample.contributed_keywords)
                threads.append(t)
                t.start()
            [t.join() for t in threads]
        self.assertLessEqual(timer.duration, self.PERIOD_SECONDS, timer.duration)

        time.sleep(self.PERIOD_SECONDS)
        threads = []
        with self.Timer() as timer:
            for sample in pyrefinebio.Experiment.get("OVER").samples:
                t = threading.Thread(target=lambda: sample.contributed_keywords)
                threads.append(t)
                t.start()
            [t.join() for t in threads]
        self.assertGreater(timer.duration, self.PERIOD_SECONDS, timer.duration)
